﻿using GalaSoft.MvvmLight;
using SciChart.Charting.Model.DataSeries;
using SciChart.Charting.Visuals;
using SciChart.Charting.Visuals.Axes;
using SciChart.Data.Model;
using SciChartPerfTester.Utilities;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;

namespace SciChartPerfTester.ViewModel
{
	public class ChartViewModel : ViewModelBase
	{
		private object _syncObject = new object();
		private Task<bool> _updateTask = null;

		private int _gapSize = 50;
		private const double _MinSourceValue = -0.50;
		private const double _MaxSourceValue = 1.50;
		private const double _SourceDelta = 0.005;
		private double _sourceValue = 0;
		private double _sourceDelta = _SourceDelta;
		private long _currentIndex;
		private bool _renderDataA = true;

		// pts per second
		private double _dataFrequency;

		public ChartViewModel(int numPointsPerSeries)
		{
			if (IsInDesignMode)
			{
				// Code runs in Blend --> create design time data.
			}
			else
			{
				// Code runs "for real"
			}

			NumPointsPerSeries = numPointsPerSeries;
		}

		private bool _isRunning;
		public bool IsRunning { get { return _isRunning; } private set { Set(nameof(IsRunning), ref _isRunning, value); } }

		private int _numPointsPerSeries;
		public int NumPointsPerSeries
		{
			get
			{
				return _numPointsPerSeries;
			}
			set
			{
				if (Set(nameof(NumPointsPerSeries), ref _numPointsPerSeries, value))
				{
					ReinitializeSeries();
					RaisePropertyChanged(nameof(XRange));
				}
			}
		}

		private int _chartTimeInMs = 5000;
		public int ChartTimeInMs
		{
			get { return _chartTimeInMs; }
			set
			{
				if (Set(nameof(ChartTimeInMs), ref _chartTimeInMs, value))
				{
					ReinitializeSeries();
					RaisePropertyChanged(nameof(XRange));
				}
			}
		}

		private XyDataSeries<long, double> _dataSeriesA;
		public XyDataSeries<long, double> DataSeriesA
		{
			get { return _dataSeriesA; }
			private set { Set(nameof(DataSeriesA), ref _dataSeriesA, value); }
		}

		private XyDataSeries<long, double> _dataSeriesB;
		public XyDataSeries<long, double> DataSeriesB
		{
			get { return _dataSeriesB; }
			private set { Set(nameof(DataSeriesB), ref _dataSeriesB, value); }
		}

		public IRange XRange
		{
			get
			{
				return new DoubleRange(0, NumPointsPerSeries);
			}
		}

		private int _timeToSleepInMs = 40;
		public int TimeToSleepInMs { get { return _timeToSleepInMs; } set { Set(nameof(TimeToSleepInMs), ref _timeToSleepInMs, value); } }

		private Visibility _diagnosticsVisibility;
		public Visibility DiagnosticsVisibility
		{
			get { return _diagnosticsVisibility; }
			set { Set(nameof(DiagnosticsVisibility), ref _diagnosticsVisibility, value); }
		}

		public void Start()
		{
			lock (_syncObject)
			{
				Log.D("Starting chart...");
				if (IsRunning)
				{
					Log.D("Attempting to start an already started VM.");
					return;
				}

				_lastRenderTime = DateTime.Now;

				_updateTask = Task.Factory.StartNew(
					() => UpdateTaskFunction(),
					TaskCreationOptions.LongRunning
				);

				IsRunning = true;
			}
		}

		public void Stop()
		{
			lock (_syncObject)
			{
				Log.D("Stopping chart...");
				if (!IsRunning)
				{
					Log.D("Attempting to stop a non-running VM.");
					return;
				}

				IsRunning = false;
				// wait at least as long as the time it should sleep before trying to dispose the Task
				if (!_updateTask.Wait(TimeSpan.FromMilliseconds(TimeToSleepInMs * 1.5)))
				{
					try
					{
						_updateTask.Dispose();
					}
					catch (Exception ex)
					{
						Log.D($"Exception on stopping: {ex}");
					}
				}
			}
		}

		private void ReinitializeSeries()
		{
			bool restartRequired = IsRunning;
			if (restartRequired)
				Stop();

			_currentIndex = 0;
			_dataFrequency = NumPointsPerSeries / (ChartTimeInMs * 0.001);

			DataSeriesA = new XyDataSeries<long, double>(NumPointsPerSeries) { FifoCapacity = NumPointsPerSeries };
			DataSeriesB = new XyDataSeries<long, double>(NumPointsPerSeries) { FifoCapacity = NumPointsPerSeries };

			// 10% of total number of points
			_gapSize = (int)Math.Round(NumPointsPerSeries / 10.0);
			DataSeriesA.AcceptsUnsortedData = !_UseUpdate;
			DataSeriesB.AcceptsUnsortedData = !_UseUpdate;

			var xVals = Enumerable.Range(0, NumPointsPerSeries + 1).Select(i => (long)i);
			var yVals = xVals.Select(i => double.NaN);

			DataSeriesA.Append(xVals, yVals);
			DataSeriesB.Append(xVals, yVals);

			//if (restartRequired)
			//Start();
		}


		private bool UpdateTaskFunction()
		{
			Log.D("Starting update function...");
			try
			{
				while (IsRunning)
				{
					Tick(DateTime.Now);
					Thread.Sleep(TimeToSleepInMs);
				}
			}
			catch (Exception ex)
			{
				Log.D("Exception caught: " + ex);
			}

			Log.D("Update function complete.");
			return true;
		}

		// used to determine exactly how many points we need to render on a given update
		// based on our desired frequency and how much time actually passed since last render
		private DateTime _lastRenderTime = DateTime.Now;
		private double _partialPoints = 0.0;

		private object _syncRoot = new object();

		public void Tick(DateTime now)
		{
			lock (_syncRoot)
			{
				var timeToRenderFor = now.Subtract(_lastRenderTime);
				var numberOfPointsToRender = _dataFrequency * timeToRenderFor.TotalSeconds;

				// keep track of the decimal portion of number of points to render
				_partialPoints += numberOfPointsToRender - (int)numberOfPointsToRender;
				int numberOfPointsToRenderAsInt = (int)numberOfPointsToRender;

				// when we have a whole point from decimal portions, add that to the number to render
				if (_partialPoints >= 1.0)
				{
					// add the integer portion of partial points
					numberOfPointsToRenderAsInt += (int)_partialPoints;

					// remove int portion and keep remaining decimals
					_partialPoints -= (int)_partialPoints;
				}

				AppendPointsWithSuspendUpdates(numberOfPointsToRenderAsInt);

				// remember the last time we rendered
				_lastRenderTime = now;
			}
		}

		public SciChartSurface SciChartSurface { get; set; }

		// different strategies for updating
		private bool _UseUpdate = true;
		private bool _UseLists = false;

		private void AppendPointsWithSuspendUpdates(int numPointsToRender)
		{
			using (var surfaceSuspender = SciChartSurface.SuspendUpdates())
			using (var aSuspender = DataSeriesA.SuspendUpdates())
			using (var bSuspender = DataSeriesB.SuspendUpdates())
			{
				AppendPoints(numPointsToRender);
			}
		}

		private void AppendPoints(int numPointsToRender)
		{
			if (_UseLists)
			{
				AppendPointsViaAppendList(numPointsToRender);
			}
			else if (_UseUpdate)
			{
				AppendPointsViaUpdate(numPointsToRender);
			}
			else
			{
				AppendPointsViaAppendPoint(numPointsToRender);
			}
		}

		private void AppendPointsViaUpdate(int numPointsToRender)
		{
			for (int j = 0; j < numPointsToRender; j++)
			{
				if (_currentIndex >= NumPointsPerSeries - 1)
				{
					_currentIndex = 0;
					_renderDataA = !_renderDataA;
				}

				double voltage = UpdateVoltage();

				try
				{
					if (_renderDataA)
					{
						DataSeriesA.Update(_currentIndex, voltage);
						DataSeriesA.Update((_currentIndex + _gapSize) % NumPointsPerSeries, double.NaN);
						DataSeriesB.Update((_currentIndex + _gapSize) % NumPointsPerSeries, double.NaN);
					}
					else
					{
						DataSeriesA.Update((_currentIndex + _gapSize) % NumPointsPerSeries, double.NaN);
						DataSeriesB.Update(_currentIndex, voltage);
						DataSeriesB.Update((_currentIndex + _gapSize) % NumPointsPerSeries, double.NaN);
					}
				}
				catch (Exception ex)
				{
					Log.D("Exception caught: " + ex);
					Log.D("_currentIndex: " + _currentIndex);
					_currentIndex = 0;
					_renderDataA = !_renderDataA;
				}
				_currentIndex++;
			}
		}

		private void AppendPointsViaAppendPoint(int numPointsToRender)
		{
			for (int j = 0; j < numPointsToRender; j++)
			{

				if (_currentIndex >= NumPointsPerSeries - 1)
				{
					_currentIndex = 0;
					_renderDataA = !_renderDataA;
				}

				// Get the next voltage and time, and append to the chart
				double voltage = UpdateVoltage();
				try
				{
					if (_renderDataA)
					{
						DataSeriesA.Append(_currentIndex, voltage);
						DataSeriesA.Append((_currentIndex + _gapSize) % NumPointsPerSeries, double.NaN);
						DataSeriesB.Append((_currentIndex + _gapSize) % NumPointsPerSeries, double.NaN);
					}
					else
					{
						DataSeriesA.Append((_currentIndex + _gapSize) % NumPointsPerSeries, double.NaN);
						DataSeriesB.Append(_currentIndex, voltage);
						DataSeriesB.Append((_currentIndex + _gapSize) % NumPointsPerSeries, double.NaN);
					}
				}
				catch (Exception ex)
				{
					Log.D("Exception caught: " + ex);
					Log.D("_currentIndex: " + _currentIndex);
					_currentIndex = 0;
					_renderDataA = !_renderDataA;
				}
				_currentIndex++;
			}
		}



		private void AppendPointsViaAppendList(int numPointsToRender)
		{
			var xVals = new List<long>();
			var yValsA = new List<double>();
			var yValsB = new List<double>();


			for (int j = 0; j < numPointsToRender; j++)
			{

				if (_currentIndex >= NumPointsPerSeries - 1)
				{
					_currentIndex = 0;
					_renderDataA = !_renderDataA;
				}

				double voltage = UpdateVoltage();

				try
				{
					xVals.Add(_currentIndex);

					if (_renderDataA)
					{
						yValsA.Add(voltage);
						yValsB.Add(double.NaN);
					}
					else
					{
						yValsA.Add(double.NaN);
						yValsB.Add(voltage);
					}
				}
				catch (Exception ex)
				{
					Log.D("Exception caught: " + ex);
					Log.D("_currentIndex: " + _currentIndex);
					_currentIndex = 0;
					_renderDataA = !_renderDataA;
				}
				_currentIndex++;
			}

			DataSeriesA.Append(xVals, yValsA);
			DataSeriesB.Append(xVals, yValsB);

		}

		private double UpdateVoltage()
		{
			double voltage = _sourceValue += _sourceDelta;
			if (_sourceValue >= _MaxSourceValue)
			{
				_sourceDelta = -1 * _SourceDelta;
			}
			else if (_sourceValue <= _MinSourceValue)
			{
				_sourceDelta = _SourceDelta;
			}
			return voltage;
		}
	}
}
