using GalaSoft.MvvmLight;
using GalaSoft.MvvmLight.Command;
using SciChart.Drawing.DirectX.Context.D3D10;
using SciChartPerfTester.Utilities;
using System;
using System.Collections.ObjectModel;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Input;
using System.Windows.Threading;

namespace SciChartPerfTester.ViewModel
{
	/// <summary>
	/// This class contains properties that the main View can data bind to.
	/// <para>
	/// Use the <strong>mvvminpc</strong> snippet to add bindable properties to this ViewModel.
	/// </para>
	/// <para>
	/// You can also use Blend to data bind with the tool's support.
	/// </para>
	/// <para>
	/// See http://www.galasoft.ch/mvvm
	/// </para>
	/// </summary>
	public class MainViewModel : ViewModelBase
	{
		private Task<bool> _updateTask = null;

		public MainViewModel()
		{
			if (IsInDesignMode)
			{
				// Code runs in Blend --> create design time data.
			}
			else
			{
				//SciChart.Core.Utility.SciChartDebugLogger.Instance.SetLogger(new Log());

				Log.D($"Direct3D10CompatibilityHelper.HasDirectX10CapableGpu: {Direct3D10CompatibilityHelper.HasDirectX10CapableGpu}");
				Log.D($"Direct3D10CompatibilityHelper.HasDirectX10RuntimeInstalled: {Direct3D10CompatibilityHelper.HasDirectX10RuntimeInstalled}");
				Log.D($"Direct3D10CompatibilityHelper.IsSupportedOperatingSystem: {Direct3D10CompatibilityHelper.IsSupportedOperatingSystem}");
				Log.D($"Direct3D10CompatibilityHelper.SupportsDirectX10: {Direct3D10CompatibilityHelper.SupportsDirectX10}");
			}

			StartChartsCommand = new RelayCommand(DoStartCharts);
			StopChartsCommand = new RelayCommand(DoStopCharts);

			Charts = new ObservableCollection<ChartViewModel>();
			InitializeCharts();
			InitializeDefaults();
		}

		private void InitializeDefaults()
		{
			NumCharts = 2;
			TimeToSleepInMs = 40;

			// 200 points per second
			ChartTimeInMs = 5000; // 5 seconds
			NumPointsPerChart = 1000;

			ShowDiagnostics = false;
			UseSingleTaskForUpdate = false;
		}

		public ICommand StartChartsCommand { get; private set; }
		public ICommand StopChartsCommand { get; private set; }

		private int _numCharts = 2;
		public int NumCharts
		{
			get { return _numCharts; }
			set { if (Set(nameof(NumCharts), ref _numCharts, value)) { InitializeCharts(); } }
		}

		private bool _showDiagnostics = true;
		public bool ShowDiagnostics
		{
			get { return _showDiagnostics; }
			set { SetOnCharts(nameof(ShowDiagnostics), ref _showDiagnostics, value, c => c.DiagnosticsVisibility = VisFromBool(value)); }
		}

		private int _numPointsPerChart;
		public int NumPointsPerChart
		{
			get { return _numPointsPerChart; }
			set { SetOnCharts(nameof(NumPointsPerChart), ref _numPointsPerChart, value, c => c.NumPointsPerSeries = value); }
		}

		private int _chartTimeInMs;
		public int ChartTimeInMs
		{
			get { return _chartTimeInMs; }
			set { SetOnCharts(nameof(ChartTimeInMs), ref _chartTimeInMs, value, c => c.ChartTimeInMs = value); }
		}

		private int _timeToSleepInMs;
		public int TimeToSleepInMs
		{
			get { return _timeToSleepInMs; }
			set { SetOnCharts(nameof(TimeToSleepInMs), ref _timeToSleepInMs, value, c => c.TimeToSleepInMs = value); }
		}

		private bool _isRunning;
		public bool IsRunning
		{
			get { return _isRunning; }
			set { if (Set(nameof(IsRunning), ref _isRunning, value)) { RaisePropertyChanged(nameof(IsNotRunning)); } }
		}

		public bool IsNotRunning { get { return !_isRunning; } }

		private bool _useSingleTaskForUpdate;

		public bool UseSingleTaskForUpdate
		{
			get { return _useSingleTaskForUpdate; }
			set { Set(nameof(UseSingleTaskForUpdate), ref _useSingleTaskForUpdate, value); }
		}

		public ObservableCollection<ChartViewModel> Charts { get; private set; }

		private void InitializeCharts()
		{
			if (_useSingleTaskForUpdate)
			{
				while (NumCharts < Charts.Count)
				{
					var chartToRemove = Charts.Last();
					chartToRemove.Stop();
					Charts.Remove(chartToRemove);
				}

				while (NumCharts > Charts.Count)
				{
					var chartToAdd = new ChartViewModel(NumPointsPerChart)
					{
						DiagnosticsVisibility = VisFromBool(ShowDiagnostics)
					};
					Charts.Add(chartToAdd);
				}
			}
			else
			{
				while (NumCharts < Charts.Count)
				{
					Charts.Remove(Charts.Last());
				}

				while (NumCharts > Charts.Count)
				{
					var chartToAdd = new ChartViewModel(NumPointsPerChart)
					{
						DiagnosticsVisibility = VisFromBool(ShowDiagnostics)
					};
					Charts.Add(chartToAdd);
				}
			}
		}
		private void SetOnCharts<T>(string propName, ref T field, T value, Action<ChartViewModel> chartSetter)
		{
			if (Set(propName, ref field, value))
			{
				DoOnAllCharts(chartSetter);
			}
		}

		private void DoOnAllCharts(Action<ChartViewModel> action)
		{
			Charts.AsParallel().ForAll(c => action(c));
		}

		private void DoStartCharts()
		{
			IsRunning = true;
			if (_useSingleTaskForUpdate)
			{

				_updateTask = Task.Factory.StartNew(() => UpdateTaskFunction(), TaskCreationOptions.LongRunning);
			}
			else
			{
				DoOnAllCharts(c => c.Start());
			}
		}

		private void DoStopCharts()
		{
			IsRunning = false;
			if (_useSingleTaskForUpdate)
			{
				Log.D("Stopping app...");
				if (!IsRunning)
				{
					Log.D("Attempting to stop a non-running app.");
					return;
				}

				// wait at least as long as the time it should sleep before trying to dispose the Task
				if (!_updateTask.Wait(TimeSpan.FromMilliseconds(TimeToSleepInMs * 1.5)))
				{
					try
					{
						_updateTask.Dispose();
					}
					catch (Exception ex)
					{
						Log.D($"Exception on stopping: {ex}");
					}
				}
			}
			else
			{
				DoOnAllCharts(c => c.Stop());
			}
		}

		private static Visibility VisFromBool(bool b)
		{
			return b ? Visibility.Visible : Visibility.Collapsed;
		}

		private bool UpdateTaskFunction()
		{
			while (IsRunning)
			{
				DateTime now = DateTime.Now;
				try
				{
					Dispatcher.CurrentDispatcher.Invoke(
						() => { foreach (var c in Charts) { c.Tick(now); } }
						, DispatcherPriority.Render
					);
					Thread.Sleep(TimeToSleepInMs);
				}
				catch (Exception ex)
				{
					Log.D("Exception: " + ex);
				}
			}

			return true;
		}
	}
}